
/* 
 * Copyright (c) 2003, Tetsuhiko Koto.
 * Licensed under the GPL.
 * 
 */

import java.util.*;
import java.util.regex.*;
import java.io.*;
import org.apache.tools.ant.*;
import org.apache.tools.ant.taskdefs.*;
import org.apache.tools.ant.util.*;
import jp.gr.java_conf.koto.notavascript.*;
import jp.gr.java_conf.koto.notavascript.impl.Syntax.ParseException;

public class MetaTextTask2 extends MatchingTask {
    private String encoding = "UTF-8";
    public void setEncoding(String encoding) {
        this.encoding = encoding;
    }
    
    private File base;
    public void setSrcdir(File dir) {
        base = dir;
    }
    
    private File destination;
    public void setDestdir(File dir) {
        destination = dir;
    }
    
    public void execute() throws BuildException {
        try {
            String[] filenames = getDirectoryScanner(base).getIncludedFiles();
            FileNameMapper mapper = new FileNameMapper() {
                public void setFrom(String from) {
                }
                public void setTo(String to) {
                }
                private String[] EMPTY = new String[0];
                public String[] mapFileName(String filename) {
                    String str = ".meta.";
                    int index = filename.indexOf(str);
                    if (index < 0)
                        return EMPTY;
                    for (;;) {
                        filename = filename.substring(0, index) + "." + filename.substring(index + str.length());
                        index = filename.indexOf(str);
                        if (index < 0)
                            break;
                    }
                    return new String[] { filename };
                }
            };
            filenames = new SourceFileScanner(this).restrict(filenames, base, destination, mapper);
            if (filenames.length > 0) {
                for (int i = 0; i < filenames.length; i++) {
                    String filename = filenames[i];
                    File src = new File(base, filename);
                    File dst = new File(destination, mapper.mapFileName(filename)[0]);
                    System.out.println("from " + src + " to " + dst);
                    new MetaText2(src, dst, getOwningTarget().getProject().getProperties(), encoding);
                }
            }
        } catch (Exception x) {
            x.printStackTrace();
            if (x instanceof BuildException) {
                throw (BuildException) x;
            } else {
                throw new BuildException(x);
            }
        } catch (AssertionError x) {
            x.printStackTrace();
            throw new BuildException(x);
        }
    }
    
    public static void main(String[] args) {
        try {
            if (args.length != 2)
                System.out.println("usage: MetaText2 <input file> <output file>");
            else
                new MetaText2(new File(args[0]), new File(args[1]), System.getProperties(), "UTF-8");
        } catch (Exception x) {
            x.printStackTrace();
        } catch (AssertionError x) {
            x.printStackTrace();
        }
        //try{System.in.read();}catch(Exception e){e.printStackTrace();}
    }
}

class MetaText2 {
    public MetaText2(File input, File output, Map properties, String encoding) throws Exception {
        Context context = NotavaScript.createContext();
        context.set("$classLoader", getClass().getClassLoader());
        context.setFinal("properties", properties);
        context.setFinal("docfile", input);
        context.setFinal("docfiledir", input.getParentFile());
        context.set("stringWrittenBy", null);
        NotavaScript.compile(context,
            "@0.3 stringWrittenBy = (writer){"
                + "buf = new StringWriter();"
                + "printer = new PrintWriter(buf);"
                + "writer(printer);"
                + "printer.close();"
                + "return buf.toString();"
            + "};").call();
        
        Reader reader = null;
        Writer writer = null;
        boolean good = false;
        try {
            reader = new BufferedReader(new InputStreamReader(new FileInputStream(input), encoding));
            writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(output), encoding));
            
            StringBuffer buffer = new StringBuffer((int) input.length());
            for (;;) {
                int ch = reader.read();
                if (ch < 0)
                    break;
                if (ch == '\u00A5')
                    ch = '\\';
                buffer.append((char) ch);
            }
            
            if (buffer.indexOf("<!--Pnuts") >= 0) {
                throw new IllegalArgumentException("'<!--Pnuts' is not supported.");
                //System.err.println(input + " : warning : contains <!--Pnuts");
            }
            
            // replace <!--notavaScript ... -->
            String head = "<!--notavaScript";
            String tail = "-->";
            int scriptStartIndex = -1;
            int scriptStartLine = -1;
            int line = 1;
            boolean pending = false;
            for (int index = 0; index < buffer.length(); index++) {
                char ch = buffer.charAt(index);
                if (pending) {
                    if (ch != '\n')
                        line++;
                    pending = false;
                }
                switch (ch) {
                case '\r':
                    pending = true;
                    break;
                case '\n':
                    line++;
                    break;
                case '<':
                case '-':
                    if (scriptStartIndex < 0) {
                        if (head.equals(buffer.substring(index, index + Math.min(head.length(), buffer.length() - index)))) {
                            scriptStartIndex = index;
                            scriptStartLine = line;
                        }
                    } else {
                        if (tail.equals(buffer.substring(index, index + Math.min(tail.length(), buffer.length() - index)))) {
                            String text = buffer.substring(scriptStartIndex + head.length(), index);
                            //System.out.println(scriptStartLine + " : " + text);
                            Object replacement;
                            try {
                                replacement = NotavaScript.compile(context, input.getPath(), text, 8, scriptStartLine, 1).call();
                            } catch (ParseException x) {
                                System.out.println(x.getGCCStylePositionString() + ": " + x.getMessage());
                                throw x;
                            }
                            if (!(replacement instanceof CharSequence))
                                throw new RuntimeException("error: the result of " + text + " is not a CharSequence; " + replacement);
                            String s = replacement.toString();
                            buffer.replace(scriptStartIndex, index + tail.length(), s);
                            index = scriptStartIndex + s.length() - 1;
                            scriptStartIndex = -1;
                        }
                    }
                    break;
                default:
                    break;
                }
            }
            
            /*
            int index = 0;
            for (;;) {
                index = buffer.indexOf(head, index);
                if (index < 0)
                    break;
                
                int lastIndex = buffer.indexOf(tail, index);
                if (lastIndex < 0)
                    throw new RuntimeException("error: " + tail + " missing");
                lastIndex += tail.length();
                
                String text = buffer.substring(index + head.length(), lastIndex - tail.length());
                Object replacement = NotavaScript.evaluate(context, text);
                if (!(replacement instanceof String))
                    throw new RuntimeException("error: the result of " + text + " is not string; " + replacement);
                
                String s = (String) replacement;
                buffer.replace(index, lastIndex, s);
                index += s.length();
            }
            */
            
            
            final String meta = ".meta.";
            // replace <... .meta. ...>
            boolean inTag = false;
            for (int i = 0; i < buffer.length(); i++) {
                char ch = buffer.charAt(i);
                if (ch == '<') {
                    inTag = true;
                } else if (ch == '>') {
                    inTag = false;
                }
                int end = Math.min(buffer.length(), i + meta.length());
                if (buffer.substring(i, end).equals(meta)) {
                    buffer.replace(i, i + meta.length(), ".");
                }
            }
            
            writer.write(embedCData(buffer.toString()));
            good = true;
        } finally {
            if (reader != null)
                reader.close();
            if (writer != null)
                writer.close();
            if (!good) {
                if (output.exists())
                    output.delete();
            }
        }
    }
    
    private String embedCData(String master) {
        StringBuffer buffer = new StringBuffer(master);
        String head = "<![CDATA[";
        String tail = "]]>";
        int index = 0;
        for (;;) {
            index = buffer.indexOf(head, index);
            if (index < 0)
                break;
            
            int lastIndex = buffer.indexOf(tail, index);
            if (lastIndex < 0)
                throw new RuntimeException("error: " + tail + " missing");
            lastIndex += tail.length();
            
            String text = buffer.substring(index + head.length(), lastIndex - tail.length());
            buffer.delete(index, lastIndex);
            for (int i = 0; i < text.length(); i++) {
                char ch = text.charAt(i);
                final String replacement;
                switch (ch) {
                case '"':
                    replacement = "&quot;";
                    break;
                case '&':
                    replacement = "&amp;";
                    break;
                case '<':
                    replacement = "&lt;";
                    break;
                case '>':
                    replacement = "&gt;";
                    break;
                default:
                    if (ch < '\u0021' || '\u007E' < ch) {
                        replacement = "&#" + (int) ch + ";";
                        break;
                    }
                case '\n':
                case '\r':
                case ' ':
                case '\t':
                    replacement = Character.toString(ch);
                    break;
                }
                buffer.insert(index, replacement);
                index += replacement.length();
            }
        }
        return buffer.toString();
    }
}

